# macOS 环境 GitHub Actions 优化指南

本文档总结了针对 py-dem-bones 项目在 macOS 环境下的 GitHub Actions 工作流优化策略。

## 主要优化内容

### 1. Homebrew 环境变量优化

在所有使用 Homebrew 的步骤中添加以下环境变量，显著提高性能：

```bash
export HOMEBREW_NO_AUTO_UPDATE=1     # 禁止自动更新
export HOMEBREW_NO_INSTALL_CLEANUP=1 # 禁止安装后清理
export HOMEBREW_NO_ANALYTICS=1       # 禁止分析数据收集
export HOMEBREW_NO_COLOR=1           # 禁止彩色输出
```

### 2. Eigen 库安装优化

- 添加缓存检查，避免重复安装
- 使用重试机制处理网络不稳定问题
- 使用 `--quiet` 参数减少输出
- 添加安装验证，确保安装成功

### 3. 缓存策略优化

- 添加 Homebrew 缓存，减少依赖下载时间
- 优化 Eigen 库缓存，避免重复安装
- 保留 pip 缓存，加速依赖安装

### 4. 依赖安装优化

- 使用清华大学镜像源加速依赖下载
- 配置 pip 重试参数，提高网络稳定性
- 使用 `--no-cache-dir` 参数避免缓存问题

### 5. 重试机制实现

- 创建 `retry_command.sh` 脚本，支持命令执行失败时自动重试
- 实现指数退避策略，每次重试等待时间翻倍
- 提供详细的日志输出，记录重试过程

## 优化效果

这些优化措施显著提高了 macOS 环境下的构建稳定性和性能：

1. **构建速度提升**：通过缓存和优化的依赖安装，减少了构建时间
2. **稳定性提高**：通过重试机制，降低了网络不稳定导致的构建失败
3. **资源消耗降低**：通过跳过不必要的步骤，减少了资源消耗

## 最佳实践建议

1. **使用缓存**：尽可能使用缓存机制，避免重复下载和安装
2. **添加重试机制**：对网络相关操作添加重试机制，提高稳定性
3. **优化 Homebrew**：使用环境变量优化 Homebrew 性能
4. **使用镜像源**：使用国内镜像源加速依赖下载
5. **跳过不必要的步骤**：通过条件判断，跳过已完成的步骤

## 后续优化方向

1. **进一步减少构建矩阵**：考虑在 macOS 环境中只测试最新的 Python 版本
2. **优化 Python 3.7 支持**：考虑在 macOS 环境中移除 Python 3.7 支持，或使用更高效的安装方式
3. **并行构建优化**：探索更多并行构建的可能性，进一步提高构建速度

## macOS 运行器资源优化

GitHub Actions 中的 macOS 运行器资源通常比 Linux 运行器更为有限，可能导致构建任务长时间等待。以下是一些解决方案：

### 1. 减少 macOS 构建矩阵

- 仅在主分支和发布分支上运行完整的 macOS 测试矩阵
- 在 Pull Request 中仅使用最新的 Python 版本进行 macOS 测试
- 使用条件表达式控制何时运行 macOS 构建：
  ```yaml
  if: ${{ github.event_name != 'pull_request' || github.base_ref == 'main' }}
  ```

### 2. 使用自托管运行器

如果有条件，可以考虑使用自托管的 macOS 运行器，这样可以：
- 避免等待 GitHub 托管的 macOS 运行器
- 自定义硬件配置，提高构建性能
- 预装所需的依赖，减少安装时间

### 3. 优化构建并发性

- 使用 `concurrency` 配置限制同时运行的 macOS 作业数量
- 优先处理关键分支的构建任务
- 取消正在运行的冗余构建，例如当有新的提交推送到同一个 PR 时

### 4. 使用缓存预热

- 定期运行构建任务以预热缓存
- 使用 `actions/cache` 的 `restore-keys` 功能利用部分匹配的缓存
- 考虑使用 GitHub Packages 或其他制品存储来缓存预构建的依赖
